if (not Grid) then return end


-- Libraries

local L = AceLibrary("AceLocale-2.2"):new("GridLayoutPlus")
local RL = AceLibrary("Roster-2.1")


-- New Grid module

GridLayoutPlus = Grid:NewModule("GridLayoutPlus", "AceConsole-2.0")

GridLayoutPlus.defaultDB = {
	dynamicLayoutEnable = true,
	layoutSorting = "GROUP",
	showPets = true,
	showTanks = nil,
	showPetSpacer = nil,
	showTankSpacer = nil,
	unitsPerColumn = 5,
	classOrder = "WARRIOR,ROGUE,DRUID,PALADIN,SHAMAN,PRIEST,MAGE,WARLOCK,HUNTER",
	petOrder = "HUNTER,WARLOCK",
	roleOrder = "MAINTANK,MAINASSIST",
	columnsBeforeHide = 20,
	warnPetsHidden = nil,
	warnTanksHidden = nil,
}


-- Locals

local playerColumns = 0
local petColumns = 0
local tankColumns = 0
local tankType = nil
local raTankList = ""

local updateNeeded = nil


-- Default layout

GridLayoutPlus.DynamicLayout = {
	defaults = {
		maxColumns = 8,
	},
}

GridLayout:AddLayout(L["DynamicLayout"], GridLayoutPlus.DynamicLayout)


-- Init

function GridLayoutPlus:OnEnable()
	local options = GridLayout.options.args
	options["dynamicLayoutEnable"] = {
		type = "toggle",
		name = L["Enable Dynamic Layout (+)"],
		desc = L["Enable dynamic layout auto-adjustments."],
		order = 107,
		get =	function()
					return GridLayoutPlus.db.profile.dynamicLayoutEnable
				end,
		set =	function(v)
					GridLayoutPlus.db.profile.dynamicLayoutEnable = v
					GridLayoutPlus:FullUpdate()
				end,
	}
	options["settings"] = {
		type = "group",
		name = L["Dynamic Settings (+)"],
		desc = L["Dynamic layout settings."],
		order = 108,
		args = {
			["layoutSorting"] = {
				type = "text",
				name = L["Layout Sorting"],
				desc = L["Choose between Group or Class sorting."],
				order = 10,
				get =	function()
							return GridLayoutPlus.db.profile.layoutSorting
						end,
				set =	function(v)
							GridLayoutPlus.db.profile.layoutSorting = v
							updateNeeded = true
							GridLayoutPlus:RosterUpdate()
						end,
				validate = {
					["GROUP"] = L["By Group"],
					["CLASS"] = L["By Class"],
				},
			},
			["showPets"] = {
				type = "toggle",
				name = L["Show Pets"],
				desc = L["Show pets in raids."],
				order = 20,
				get =	function()
							return GridLayoutPlus.db.profile.showPets
						end,
				set =	function(v)
							GridLayoutPlus.db.profile.showPets = v
							dynamicUpdateNeeded = true
							GridLayoutPlus:RosterUpdate()
						end,
			},
			["showPetSpacer"] = {
				type = "toggle",
				name = L["Show Pet Spacer"],
				desc = L["Show a spacer column between players and pets."],
				order = 30,
				get =	function()
							return GridLayoutPlus.db.profile.showPetSpacer
						end,
				set =	function(v)
							GridLayoutPlus.db.profile.showPetSpacer = v
							dynamicUpdateNeeded = true
							GridLayoutPlus:LayoutUpdate()
						end,
			},
			["showTanks"] = {
				type = "toggle",
				name = L["Show Tanks"],
				desc = L["Show tanks in raids."],
				order = 40,
				get =	function()
							return GridLayoutPlus.db.profile.showTanks
						end,
				set =	function(v)
							GridLayoutPlus.db.profile.showTanks = v
							GridLayoutPlus:FullUpdate()
						end,
			},
			["showTankSpacer"] = {
				type = "toggle",
				name = L["Show Tank Spacer"],
				desc = L["Show a spacer column between tanks and players."],
				order = 50,
				get =	function()
							return GridLayoutPlus.db.profile.showTankSpacer
						end,
				set =	function(v)
							GridLayoutPlus.db.profile.showTankSpacer = v
							dynamicUpdateNeeded = true
							GridLayoutPlus:LayoutUpdate()
						end,
			},
			["unitsPerColumn"] = {
				type = "range",
				name = L["Units Per Column"],
				desc = L["How many units to show in a single column."],
				order = 60,
				min = 5,
				max = 40,
				step = 5,
				get =	function()
							return GridLayoutPlus.db.profile.unitsPerColumn
						end,
				set =	function(v)
							GridLayoutPlus.db.profile.unitsPerColumn = v
							dynamicUpdateNeeded = true
							GridLayoutPlus:FullUpdate()
						end,
			},
			["classOrder"] = {
				type = "text",
				name = L["Class Order"],
				desc = L["Order in which player units appear."],
				usage = L["Order in which player units appear."],
				order = 70,
				get =	function()
							return GridLayoutPlus.db.profile.classOrder
						end,
				set =	function(v)
							GridLayoutPlus.db.profile.classOrder = v
							dynamicUpdateNeeded = true
							GridLayoutPlus:RosterUpdate()
						end,
			},
			["petOrder"] = {
				type = "text",
				name = L["Pet Order/Filter"],
				desc = L["Order in which pet units appear (also filters)."],
				usage = L["Order in which pet units appear (also filters)."],
				order = 80,
				get =	function()
							return GridLayoutPlus.db.profile.petOrder
						end,
				set =	function(v)
							GridLayoutPlus.db.profile.petOrder = v
							dynamicUpdateNeeded = true
							GridLayoutPlus:RosterUpdate()
						end,
			},
			["roleOrder"] = {
				type = "text",
				name = L["Blizz-tank Order/Filter"],
				desc = L["Order in which Blizzard tanks appear (also filters)."],
				usage = L["Order in which Blizzard tanks appear (also filters)."],
				order = 90,
				get =	function()
							return GridLayoutPlus.db.profile.roleOrder
						end,
				set =	function(v)
							GridLayoutPlus.db.profile.roleOrder = v
							dynamicUpdateNeeded = true
							GridLayoutPlus:RosterUpdate()
						end,
			},
			["advanced"] = {
				type = "group",
				name = L["Advanced"],
				desc = L["Advanced options for dynamic layouts."],
				order = -1,
				args = {
					["columnsBeforeHide"] = {
						type = "range",
						name = L["Columns Before Hiding"],
						desc = L["Max Grid columns before hiding tanks/pets."],
						order = 10,
						min = 5,
						max = 20,
						step = 1,
						get =	function()
									return GridLayoutPlus.db.profile.columnsBeforeHide
								end,
						set =	function(v)
									GridLayoutPlus.db.profile.columnsBeforeHide = v
									dynamicUpdateNeeded = true
									GridLayoutPlus:LayoutUpdate()
								end,
					},
					["warnPetsHidden"] = {
						type = "toggle",
						name = L["Warn if Pets are Hidden"],
						desc = L["Show warning when pet units are forced to be hidden."],
						order = 20,
						get =	function()
									return GridLayoutPlus.db.profile.warnPetsHidden
								end,
						set =	function(v)
									GridLayoutPlus.db.profile.warnPetsHidden = v
									dynamicUpdateNeeded = true
									GridLayoutPlus:LayoutUpdate()
								end,
					},
					["warnTanksHidden"] = {
						type = "toggle",
						name = L["Warn if Tanks are Hidden"],
						desc = L["Show warning when tank units are forced to be hidden."],
						order = 30,
						get =	function()
									return GridLayoutPlus.db.profile.warnTanksHidden
								end,
						set =	function(v)
									GridLayoutPlus.db.profile.warnTanksHidden = v
									dynamicUpdateNeeded = true
									GridLayoutPlus:LayoutUpdate()
								end,
					},
				},
			},
		},
	},
	self:RegisterEvent("Grid_JoinedRaid", "FullUpdate")
	self:RegisterEvent("Grid_JoinedBattleground", "FullUpdate")
	self:RegisterBucketEvent({"Grid_UnitJoined","Grid_UnitLeft","Grid_UnitChanged"}, 2, "RosterUpdate")
	if oRA then
		self:RegisterEvent("oRA_MainTankUpdate", "RATanksUpdate")
		self:RegisterEvent("oRA_JoinedRaid", "RATanksUpdate")
	elseif CT_RA_MainTanks then
		hooksecurefunc("CT_RAOptions_UpdateMTs", function() self:RATanksUpdate() end)
	end
	self:RegisterEvent("GridLayoutPlus_LayoutUpdate", "LayoutUpdate")
	self:ScheduleRepeatingEvent("GridLayoutPlus_LayoutUpdate", 2)
	self:FullUpdate()
end


-- Event handlers

function GridLayoutPlus:FullUpdate()
	if (GetNumRaidMembers() < 1) then return end
	self:RATanksUpdate()
	self:RosterUpdate()
	updateNeeded = true
	self:LayoutUpdate()
end

function GridLayoutPlus:UnitUpdate(name)
	if (GetNumRaidMembers() < 1) then return end
	local u = RL:GetUnitObjectFromName(name)
	if u and (u.class ~= "PET") then
		if (GridLayoutPlus.db.profile.layoutSorting == "GROUP")
				or (GridLayoutPlus.db.profile.showPets and string.find(GridLayoutPlus.db.profile.petOrder, u.class))
				or (GridLayoutPlus.db.profile.showTanks and (tankType ~= "RA")) then
			self:RosterUpdate()
		end
	end
end

function GridLayoutPlus:RATanksUpdate()
	local ratanktable = nil
	local numberoftanks = 0
	raTankList = ""
	if GridLayoutPlus.db.profile.showTanks and oRA and oRA.maintanktable then
		ratanktable = oRA.maintanktable
	elseif GridLayoutPlus.db.profile.showTanks and CT_RA_MainTanks then
		ratanktable = CT_RA_MainTanks
	end
	if ratanktable then
		for key, val in pairs(ratanktable) do
			local unitid = RL:GetUnitIDFromName(val)
			if unitid and UnitExists(unitid) and UnitPlayerOrPetInRaid(unitid) then
				numberoftanks = numberoftanks + 1
				raTankList = raTankList..val..","
			end
		end
	end
	tankColumns = floor((numberoftanks / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))
	if (tankColumns > 0) then
		tankType = "RA"
	elseif (tankType == "RA") then
		tankType = nil
		raTankList = ""
	end
	updateNeeded = true
end

function GridLayoutPlus:RosterUpdate()
	if (GetNumRaidMembers() < 1) then return end
	local maxsubgroup = 0
	local numberofpotentialpets = 0
	local numberoftanks = 0
	for u in RL:IterateRoster(nil) do
		if (GridLayoutPlus.db.profile.layoutSorting == "GROUP") and (u.subgroup > maxsubgroup) then
			maxsubgroup = u.subgroup
		end
		if GridLayoutPlus.db.profile.showPets and string.find(GridLayoutPlus.db.profile.petOrder, u.class) then
			numberofpotentialpets = numberofpotentialpets + 1
		end
		if GridLayoutPlus.db.profile.showTanks and (tankType ~= "RA") and string.find(GridLayoutPlus.db.profile.roleOrder, u.role) then
			numberoftanks = numberoftanks + 1
		end
	end
	if (GridLayoutPlus.db.profile.layoutSorting == "GROUP") and (playerColumns ~= floor((maxsubgroup * 5 / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))) then
		playerColumns = floor((maxsubgroup * 5 / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))
		updateNeeded = true
	elseif (playerColumns ~= floor((GetNumRaidMembers() / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))) then
		playerColumns = floor((GetNumRaidMembers() / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))
		updateNeeded = true
	end
	if (petColumns ~= floor((numberofpotentialpets / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))) then
		petColumns = floor((numberofpotentialpets / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))
		updateNeeded = true
	end
	if (tankColumns ~= floor((numberoftanks / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))) then
		tankColumns = floor((numberoftanks / GridLayoutPlus.db.profile.unitsPerColumn) + ((GridLayoutPlus.db.profile.unitsPerColumn - 1) / GridLayoutPlus.db.profile.unitsPerColumn))
		if (tankColumns > 0) then
			tankType = "BLIZZ"
		elseif (tankType == "BLIZZ") then
			tankType = nil
		end
		updateNeeded = true
	end
end


-- Update layout

function GridLayoutPlus:LayoutUpdate()
	if (updateNeeded) then
		if GridLayoutPlus.db.profile.dynamicLayoutEnable and (GetNumRaidMembers() > 0) and (not InCombatLockdown()) then
			local canfitpets = nil
			local canfittanks = nil
			local groupindex = 1
			local petSpacer = 0
			local tankSpacer = 0
			if GridLayoutPlus.db.profile.showPets and (petColumns > 0) then
				petSpacer = (GridLayoutPlus.db.profile.showPetSpacer and 1 or 0)
			end
			if GridLayoutPlus.db.profile.showTanks and (tankColumns > 0) then
				tankSpacer = (GridLayoutPlus.db.profile.showTankSpacer and 1 or 0)
			end
			-- Can fit pets / tanks
			if (GridLayoutPlus.db.profile.columnsBeforeHide >= (playerColumns + petColumns + petSpacer + tankColumns + tankSpacer)) then
				canfitpets = true
				canfittanks = true
			elseif (GridLayoutPlus.db.profile.columnsBeforeHide >= (playerColumns + petColumns + petSpacer)) then
				canfitpets = true
			end
			-- Alter layout
			if GridLayoutPlus.db.profile.showTanks and canfittanks and (tankType == "RA") then
				GridLayoutPlus.DynamicLayout[groupindex] = {
					nameList = raTankList,
					sortMethod = "INDEX",
					unitsPerColumn = GridLayoutPlus.db.profile.unitsPerColumn,
				}
				groupindex = groupindex + 1
			elseif GridLayoutPlus.db.profile.showTanks and canfittanks and (tankType == "BLIZZ") then
				GridLayoutPlus.DynamicLayout[groupindex] = {
					groupFilter = GridLayoutPlus.db.profile.roleOrder,
					groupingOrder = GridLayoutPlus.db.profile.roleOrder,
					groupBy = "ROLE",
					unitsPerColumn = GridLayoutPlus.db.profile.unitsPerColumn,
				}
				groupindex = groupindex + 1
			end
			if GridLayoutPlus.db.profile.showTanks and canfittanks and GridLayoutPlus.db.profile.showTankSpacer then
				GridLayoutPlus.DynamicLayout[groupindex] = {
					groupFilter = ""
				}
				groupindex = groupindex + 1
			end
			if (GridLayoutPlus.db.profile.layoutSorting == "GROUP") then
				local groupnumber = 1
				local strgroupfilter
				for i = 1, playerColumns do
					strgroupfilter = ""
					for j = groupnumber,  (groupnumber + (GridLayoutPlus.db.profile.unitsPerColumn / 5) - 1) do
						strgroupfilter = strgroupfilter .. j .. ","
						groupnumber = j
					end
					GridLayoutPlus.DynamicLayout[groupindex] = {
						groupFilter = strgroupfilter,
						groupingOrder = GridLayoutPlus.db.profile.classOrder,
						groupBy = "CLASS",
						unitsPerColumn = GridLayoutPlus.db.profile.unitsPerColumn,
					}
					groupnumber = groupnumber + 1
					groupindex = groupindex + 1
				end
			else
				GridLayoutPlus.DynamicLayout[groupindex] = {
					groupFilter = GridLayoutPlus.db.profile.classOrder,
					groupingOrder = GridLayoutPlus.db.profile.classOrder,
					groupBy = "CLASS",
					unitsPerColumn = GridLayoutPlus.db.profile.unitsPerColumn,
				}
				groupindex = groupindex + 1
			end
			if GridLayoutPlus.db.profile.showPets and canfitpets and GridLayoutPlus.db.profile.showPetSpacer then
				GridLayoutPlus.DynamicLayout[groupindex] = {
					groupFilter = ""
				}
				groupindex = groupindex + 1
			end
			if GridLayoutPlus.db.profile.showPets and canfitpets and (petColumns > 0) then
				GridLayoutPlus.DynamicLayout[groupindex] = {
					isPetGroup = true,
					groupFilter = GridLayoutPlus.db.profile.petOrder,
					groupingOrder = GridLayoutPlus.db.profile.petOrder,
					groupBy = "CLASS",
					filterOnPet = true,
					unitsPerColumn = GridLayoutPlus.db.profile.unitsPerColumn,
				}
				groupindex = groupindex + 1
			end
			for i = groupindex, 10 do
				GridLayoutPlus.DynamicLayout[i] = nil
			end
			-- Refresh layout
			GridLayout:LoadLayout("DynamicLayout")
			-- Warn if hidden units
			if GridLayoutPlus.db.profile.warnPetsHidden and GridLayoutPlus.db.profile.showPets and (petColumns > 0) and (not canfitpets) then
				self:Print(L["Your maxColumnsBeforeHide setting is causing pet units to be hidden"])
			end
			if GridLayoutPlus.db.profile.warnTanksHidden and GridLayoutPlus.db.profile.showTanks and (tankColumns > 0) and (not canfittanks) then
				self:Print(L["Your maxColumnsBeforeHide setting is causing tank units to be hidden"])
			end
			updateNeeded = nil
		elseif GridLayoutPlus.db.profile.dynamicLayoutEnable and (GetNumRaidMembers() > 0) and InCombatLockdown() then
			self:ScheduleLeaveCombatAction(self, "RosterUpdate")
		end
	end
end
